module Twitter.Http
where

import qualified Data.ByteString.Lazy as B
import qualified Data.ByteString.Char8 as BC
import qualified Data.Conduit as C
import Data.Conduit.Binary (sinkLbs)
import Network.HTTP.Conduit
import Web.Authenticate.OAuth (
  OAuth(..),
  Credential,
  newCredential,
  newOAuth,
  signOAuth)

-- |The API URL of username's timeline.
--
-- See,
--
--   <https://dev.twitter.com/docs/api/1.1/get/statuses/user_timeline>
--
user_timeline_url :: String -> String
user_timeline_url username =
  concat [ "https://api.twitter.com/",
           "1.1/",
           "statuses/",
           "user_timeline.json?",
           "screen_name=",
           username,
           "&include_rts=true&",
           "count=10" ]

status_url :: Integer -> String
status_url status_id =
  concat [ "https://api.twitter.com/",
           "1.1/",
           "statuses/",
           "show.json?id=",
           (show status_id) ]

-- | Given username's last status id, constructs the API URL for
--   username's new statuses. Essentially, 'user_timeline_url' with a
--   "since_id" parameter tacked on.
user_new_statuses_url :: String -> Integer -> String
user_new_statuses_url username last_status_id =
  concat [ user_timeline_url username,
           "&since_id=" ++ (show last_status_id) ]


get_status :: Integer -> IO B.ByteString
get_status status_id = do
  let uri = (status_url status_id)
  status <- (http_get uri)
  return status


-- | Return's username's timeline.
get_user_timeline :: String -> IO B.ByteString
get_user_timeline username = do
  let uri = (user_timeline_url username)
  timeline <- (http_get uri)
  return timeline


-- | Returns the JSON representing all of username's statuses that are
--   newer than last_status_id.
get_user_new_statuses :: String -> Integer -> IO B.ByteString
get_user_new_statuses username last_status_id = do
  let uri = (user_new_statuses_url username last_status_id)
  new_statuses <- (http_get uri)
  return new_statuses


-- | Retrieve a URL, or crash.
http_get :: String -> IO B.ByteString
http_get url = do
  manager <- newManager def
  request <- parseUrl url

  C.runResourceT $ do
    signed_request <- signOAuth oauth credential request
    response <- http signed_request manager
    responseBody response C.$$+- sinkLbs

  where
    consumer_key = BC.pack ""
    consumer_secret = BC.pack ""
    access_token = BC.pack ""
    access_secret = BC.pack ""

    oauth :: OAuth
    oauth = newOAuth {
              oauthConsumerKey = consumer_key,
              oauthConsumerSecret = consumer_secret
            }

    credential :: Credential
    credential = newCredential access_token access_secret
